import codecs
import encodings
import collections
import hashlib

charsets_ = set(encodings.aliases.aliases.values())

def detect_charset(binary):
    '''自動的に文字codeを判別する。'''

    candidates = []
    for charset in charsets_:
        try:
            text = binary.decode(charset, errors='strict')
        except LookupError as raiz:
            if raiz.args[0].startswith('unknown encoding:') or\
                False:
                continue
            print('raiz =', raiz)
            print('raiz.args =')
            print(raiz.args)
            raise raiz
        except UnicodeDecodeError as raiz:
            if raiz.args[-1] in (
                   'illegal multibyte sequence',
                   'character maps to <undefined>',
                   'codepoint not in range(0x110000)',
                   'invalid start byte',
                   'unexpected special character',
                   'ordinal not in range(128)',
                   'truncated data',
               ):
                continue
            raise raiz

        info = codecs.lookup(charset)
        encoder = info.streamreader

        try:
            encoder.encode(binary, 'strict')
        except TypeError as raiz:
            if raiz.args[0] == \
                "Can't convert 'bytes' object to str implicitly":
                continue
            raise raiz
        except NotImplementedError as raiz:
            if charset != 'utf_8':
                continue

        len_text = len(text)
        candidate = (charset, len_text)
        candidates.append(candidate)
    if not candidates:
        raise RuntimeError('{} has unkown codec.'.format(name))
    text = None

    candidates.sort(key=lambda x: x[1])
    len_min = candidates[0][1]
    i = 0
    for candidate in candidates:
        if candidate[1] != len_min:
            break
        i += 1
    candidates = candidates[:i]
    dL = collections.defaultdict(list)
    for candidate in candidates:
      # print('candidate =')
      # print(candidate)
        charset, len_text = candidate
        text = binary.decode(charset)
      # print('len(text) =', len(text))
      # print('text =')
      # print(text)
      # digest = hashlib.sha1(text.encode(charset)).digest()
        digest = hashlib.sha1(text.encode(charset)).hexdigest()
        dL[digest].append(charset)
        lines = text.split('\n')
    lines = None

    majority = 0
    for digest, charsets in dL.items():
      # print('digest = {0}, len() = {1}'.format(digest[:4], len(charsets)))
        len_charsets = len(charsets)
        if len_charsets > majority:
            majority = len_charsets
            candidate_charsets = charsets
    candidate_top = candidate_charsets[0]
  # print('candidate_charsets =', candidate_charsets)
  # print('candidate_top =', candidate_top)

    return candidate_top

if __name__ == '__main__':
    name_and_codecs = (
        ('konnichiha.EUC-JP.py', 'euc-jp'),
        ('konnichiha.ISO-2022-JP.py', 'iso-2022-jp'),
        ('konnichiha.Shift_JIS.py', 'shift_jis'),
        ('konnichiha.UTF-8.py', 'utf-8'),
    )

    print('use open() with encoding ========================================')
    for name, codec in name_and_codecs:
        with open(name, 'r', encoding=codec) as fp:
            print('name =', name)
            print('encoding =', fp.encoding)
            print(fp.readlines()[-1])
            print()

    print('use codecs.open() with encoding =================================')
    for name, codec in name_and_codecs:
        with codecs.open(name, 'r', encoding=codec) as fp:
            print('name =', name)
            print('encoding =', fp.encoding)
            print(fp.readlines()[-1])
            print()

    print('use open() without encoding =====================================')
    for name, codec in name_and_codecs:
        print('name =', name)
        with open(name, 'rb') as fp:
            binary = fp.read()
        charset = detect_charset(binary)
        print('detect_charset() =', charset)
        with codecs.open(name, 'r', encoding=charset) as fp:
            print(fp.readlines()[-1])
        print()
